#!/usr/bin/env python
import logging
from os.path import abspath, dirname

from dateutil.relativedelta import relativedelta
from edc_constants.constants import IGNORE
from edc_test_utils import DefaultTestSettings, func_main
from edc_utils import get_utcnow

app_name = "edc_pharmacy"
base_dir = dirname(abspath(__file__))

project_settings = DefaultTestSettings(
    calling_file=__file__,
    EDC_NAVBAR_VERIFY_ON_LOAD=IGNORE,
    EDC_AUTH_CODENAMES_WARN_ONLY=True,
    EDC_AUTH_SKIP_SITE_AUTHS=True,
    EDC_AUTH_SKIP_AUTH_UPDATER=True,
    BASE_DIR=base_dir,
    APP_NAME=app_name,
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=(
        get_utcnow().replace(microsecond=0, second=0, minute=0, hour=0)
        - relativedelta(years=6)
    ),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=(
        get_utcnow().replace(microsecond=999999, second=59, minute=59, hour=11)
        + relativedelta(years=6)
    ),
    add_dashboard_middleware=True,
    add_lab_dashboard_middleware=True,
    excluded_apps=["edc_adverse_event.apps.AppConfig", "adverse_event_app.apps.AppConfig"],
).settings


def main():
    func_main(project_settings, f"{app_name}.tests")


if __name__ == "__main__":
    logging.basicConfig()
    main()



from __future__ import annotations

from datetime import datetime

from dateutil.relativedelta import relativedelta


def adjust_previous_end_datetime(
    previous_obj,
    refill_start_datetime: datetime,
    user_modified: str | None = None,
    modified: str | None = None,
) -> None:
    if previous_obj.refill_end_datetime != refill_start_datetime - relativedelta(minutes=1):
        previous_obj.refill_end_datetime = refill_start_datetime - relativedelta(minutes=1)
        previous_obj.number_of_days = (
            previous_obj.refill_end_datetime - previous_obj.refill_start_datetime
        ).days
        previous_obj.user_modified = user_modified or previous_obj.user_modified
        previous_obj.modified = modified or previous_obj.modified
        previous_obj.save_base(
            update_fields=[
                "refill_end_datetime",
                "number_of_days",
                "modified",
                "user_modified",
            ]
        )



from .crf import StudyMedicationModelFormMixin
from .dispensing_history_form import (
    DispensingHistoryForm,
    DispensingHistoryReadonlyForm,
)
from .dosage_guideline_form import DosageGuidelineForm
from .formulation_form import FormulationForm
from .labels_form import LabelsForm
from .medication_form import MedicationForm
from .medication_lot_form import MedicationLotForm
from .order_form import OrderForm
from .product_form import ProductForm
from .return_history_form import ReturnHistoryForm
from .rx_form import RxForm
from .rx_refill_form import RxRefillForm
from .stock_create_labels_form import StockCreateLabelsForm
from .stock_form import StockForm
from .stock_receiving_form import StockReceivingForm



from dateutil.relativedelta import relativedelta
from django.test import TestCase
from edc_registration.models import RegisteredSubject
from edc_utils import get_utcnow

from edc_pharmacy.dispense import DispenseError
from edc_pharmacy.models import (
    DispensingHistory,
    DosageGuideline,
    Formulation,
    FormulationType,
    FrequencyUnits,
    Medication,
    Route,
    Rx,
    RxRefill,
    Units,
)
from edc_pharmacy.refill import RefillCreator


class TestDispense(TestCase):
    def setUp(self):
        self.subject_identifier = "12345"

        RegisteredSubject.objects.create(subject_identifier="12345")

        self.medication = Medication.objects.create(
            name="flucytosine",
            display_name="Flucytosine",
        )

        self.formulation = Formulation.objects.create(
            medication=self.medication,
            strength=500,
            units=Units.objects.get(name="mg"),
            route=Route.objects.get(display_name="Oral"),
            formulation_type=FormulationType.objects.get(display_name__iexact="Tablet"),
        )

        self.dosage_guideline = DosageGuideline.objects.create(
            medication=self.medication,
            dose_per_kg=100,
            dose_units=Units.objects.get(name="mg"),
            frequency=1,
            frequency_units=FrequencyUnits.objects.get(name="day"),
        )

        self.rx = Rx.objects.create(
            subject_identifier=self.subject_identifier,
            weight_in_kgs=40,
            report_datetime=get_utcnow(),
        )
        self.rx.medications.add(self.medication)

    def test_dispense(self):
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            refill_start_datetime=get_utcnow(),
            refill_end_datetime=get_utcnow() + relativedelta(days=7),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
        )
        self.assertEqual(refill_creator.rx_refill.total, 56.0)
        self.assertEqual(refill_creator.rx_refill.remaining, 56.0)

    def test_dispense_many(self):
        rx_refill = RxRefill.objects.create(
            rx=self.rx,
            formulation=self.formulation,
            dosage_guideline=self.dosage_guideline,
            refill_start_datetime=get_utcnow(),
            refill_end_datetime=get_utcnow() + relativedelta(days=7),
            weight_in_kgs=50,
        )
        dispensed = 0
        self.assertEqual(rx_refill.total, 70.0)
        for amount in [8, 8, 8]:
            dispensed += amount
            obj = DispensingHistory.objects.create(
                rx_refill=rx_refill,
                dispensed=8,
            )
            self.assertEqual(obj.dispensed, 8)
            rx_refill = RxRefill.objects.get(id=rx_refill.id)
            self.assertEqual(rx_refill.remaining, 70 - dispensed)

    def test_attempt_to_over_dispense(self):
        rx_refill = RxRefill.objects.create(
            rx=self.rx,
            formulation=self.formulation,
            dosage_guideline=self.dosage_guideline,
            refill_start_datetime=get_utcnow(),
            refill_end_datetime=get_utcnow() + relativedelta(days=8),
            weight_in_kgs=45,
        )
        dispensed = 0
        self.assertEqual(rx_refill.total, 72.0)
        for amount in [8, 8, 8, 8, 8, 8, 8, 8]:
            dispensed += amount
            obj = DispensingHistory.objects.create(
                rx_refill=rx_refill,
                dispensed=8,
            )
            self.assertEqual(obj.dispensed, 8)
            rx_refill = RxRefill.objects.get(id=rx_refill.id)
            self.assertEqual(rx_refill.remaining, 72 - dispensed)
        rx_refill = RxRefill.objects.get(id=rx_refill.id)
        self.assertEqual(rx_refill.remaining, 8)
        self.assertRaises(
            DispenseError,
            DispensingHistory.objects.create,
            rx_refill=rx_refill,
            dispensed=16,
        )



from datetime import datetime
from zoneinfo import ZoneInfo

from dateutil.relativedelta import relativedelta
from django.db.models.signals import pre_save
from django.test import TestCase, override_settings
from edc_appointment.constants import INCOMPLETE_APPT
from edc_appointment.creators import AppointmentsCreator, UnscheduledAppointmentCreator
from edc_appointment.models import Appointment
from edc_appointment.tests.helper import Helper
from edc_appointment.utils import get_next_appointment
from edc_consent.tests.consent_test_utils import consent_object_factory
from edc_constants.constants import YES
from edc_facility import import_holidays
from edc_protocol import Protocol
from edc_registration.models import RegisteredSubject
from edc_utils import get_utcnow
from edc_visit_schedule import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED, UNSCHEDULED

from edc_pharmacy.exceptions import NextStudyMedicationError, StudyMedicationError
from edc_pharmacy.models import (
    DosageGuideline,
    Formulation,
    FormulationType,
    FrequencyUnits,
    Medication,
    Route,
    Rx,
    RxRefill,
    Units,
)

from ..forms import StudyMedicationForm
from ..models import StudyMedication, SubjectVisit
from ..visit_schedule import schedule, visit_schedule


@override_settings(SUBJECT_CONSENT_MODEL="edc_pharmacy.subjectconsent")
class TestMedicationCrf(TestCase):

    helper_cls = Helper

    @classmethod
    def setUpTestData(cls):
        import_holidays()
        pre_save.disconnect(dispatch_uid="requires_consent_on_pre_save")

    def setUp(self) -> None:
        site_visit_schedules._registry = {}
        site_visit_schedules.loaded = False

        site_visit_schedules.register(visit_schedule)
        self.subject_identifier = "12345"
        self.registration_datetime = get_utcnow() - relativedelta(years=5)
        RegisteredSubject.objects.create(
            subject_identifier=self.subject_identifier,
            registration_datetime=self.registration_datetime,
            consent_datetime=self.registration_datetime,
        )
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            now=get_utcnow() - relativedelta(years=6),
        )
        self.helper.consent_and_put_on_schedule(
            subject_identifier=self.subject_identifier,
            visit_schedule_name="visit_schedule",
            schedule_name="schedule",
        )
        creator = AppointmentsCreator(
            subject_identifier=self.subject_identifier,
            visit_schedule=visit_schedule,
            schedule=schedule,
            report_datetime=self.registration_datetime,
        )
        creator.create_appointments(base_appt_datetime=self.registration_datetime)

        self.assertGreater(
            Appointment.objects.filter(subject_identifier=self.subject_identifier).count(),
            0,
        )

        self.medication = Medication.objects.create(
            name="Flucytosine",
        )

        self.formulation = Formulation.objects.create(
            medication=self.medication,
            strength=500,
            units=Units.objects.get(name="mg"),
            route=Route.objects.get(display_name="Oral"),
            formulation_type=FormulationType.objects.get(display_name__iexact="Tablet"),
        )

        self.dosage_guideline_100 = DosageGuideline.objects.create(
            medication=self.medication,
            dose_per_kg=100,
            dose_units=Units.objects.get(name="mg"),
            frequency=1,
            frequency_units=FrequencyUnits.objects.get(name="day"),
        )

        self.dosage_guideline_200 = DosageGuideline.objects.create(
            medication=self.medication,
            dose_per_kg=100,
            dose_units=Units.objects.get(name="mg"),
            frequency=2,
            frequency_units=FrequencyUnits.objects.get(name="day"),
        )

        self.rx = Rx.objects.create(
            subject_identifier=self.subject_identifier,
            weight_in_kgs=40,
            report_datetime=self.registration_datetime,
            rx_date=self.registration_datetime.date(),
        )
        self.rx.medications.add(self.medication)

    def test_ok(self):
        appointment = Appointment.objects.all().order_by("timepoint")[0]
        subject_visit = SubjectVisit.objects.create(
            appointment=appointment,
            report_datetime=appointment.appt_datetime,
            reason=SCHEDULED,
        )

        obj = StudyMedication(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            refill_start_datetime=subject_visit.report_datetime,
            refill_end_datetime=get_next_appointment(
                appointment, include_interim=True
            ).appt_datetime,
            dosage_guideline=self.dosage_guideline_100,
            formulation=self.formulation,
        )
        obj.save()

        # calc num of days until next visit
        number_of_days = (
            get_next_appointment(
                obj.subject_visit.appointment, include_interim=True
            ).appt_datetime
            - obj.subject_visit.appointment.appt_datetime
        ).days

        self.assertIsNotNone(obj.number_of_days)
        self.assertEqual(obj.number_of_days, number_of_days)
        self.assertGreater(obj.number_of_days, 0)

    def test_refill_before_rx(self):
        appointment = Appointment.objects.all().order_by("timepoint")[0]
        subject_visit = SubjectVisit.objects.create(
            appointment=appointment,
            report_datetime=appointment.appt_datetime,
            reason=SCHEDULED,
        )

        obj = StudyMedication(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            refill_start_datetime=datetime(
                self.rx.rx_date.year, self.rx.rx_date.month, self.rx.rx_date.day, 0, 0, 0
            ).astimezone(ZoneInfo("UTC"))
            - relativedelta(years=1),
            refill_end_datetime=get_next_appointment(
                appointment, include_interim=True
            ).appt_datetime,
            dosage_guideline=self.dosage_guideline_100,
            formulation=self.formulation,
        )
        with self.assertRaises(StudyMedicationError):
            obj.save()

    def test_refill_for_expired_rx(self):
        appointment = Appointment.objects.all().order_by("timepoint")[0]
        subject_visit = SubjectVisit.objects.create(
            appointment=appointment,
            report_datetime=appointment.appt_datetime,
            reason=SCHEDULED,
        )
        self.rx.rx_expiration_date = subject_visit.report_datetime.date()
        self.rx.save()
        self.rx.refresh_from_db()

        obj = StudyMedication(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            refill_start_datetime=subject_visit.report_datetime + relativedelta(years=1),
            refill_end_datetime=get_next_appointment(
                subject_visit.appointment, include_interim=True
            ).appt_datetime
            + relativedelta(years=1),
            dosage_guideline=self.dosage_guideline_100,
            formulation=self.formulation,
        )
        with self.assertRaises(StudyMedicationError):
            obj.save()

    def test_for_each_appt_creates_rxrefill_thru_studymedication(self):
        """Create one refill per appointment.

        Last appt does not get a refill
        """
        for appointment in Appointment.objects.all().order_by("timepoint"):
            subject_visit = SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=appointment.appt_datetime,
                reason=SCHEDULED,
            )
            if appointment.next:
                StudyMedication.objects.create(
                    subject_visit=subject_visit,
                    report_datetime=subject_visit.report_datetime,
                    refill_start_datetime=subject_visit.report_datetime,
                    refill_end_datetime=get_next_appointment(
                        appointment, include_interim=True
                    ).appt_datetime,
                    dosage_guideline=self.dosage_guideline_100,
                    formulation=self.formulation,
                )
        self.assertEqual(
            StudyMedication.objects.all().count(), Appointment.objects.all().count() - 1
        )
        self.assertEqual(RxRefill.objects.all().count(), Appointment.objects.all().count() - 1)

    def test_rx_refill_start_datetimes_are_greater(self):
        for appointment in Appointment.objects.all().order_by("timepoint"):
            subject_visit = SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=appointment.appt_datetime,
                reason=SCHEDULED,
            )
            if appointment.next:
                StudyMedication.objects.create(
                    subject_visit=subject_visit,
                    report_datetime=subject_visit.report_datetime,
                    refill_start_datetime=subject_visit.report_datetime,
                    refill_end_datetime=get_next_appointment(
                        appointment, include_interim=True
                    ).appt_datetime,
                    dosage_guideline=self.dosage_guideline_100,
                    formulation=self.formulation,
                )

        # check dates
        for obj in RxRefill.objects.all().order_by("refill_start_datetime"):
            self.assertLess(obj.refill_start_datetime, obj.refill_end_datetime)

        refill_start_datetimes = [
            obj.refill_start_datetime
            for obj in RxRefill.objects.all().order_by("refill_start_datetime")
        ]
        last_dt = None
        for dt in refill_start_datetimes:
            if not last_dt:
                last_dt = dt
                continue
            self.assertGreater(dt, last_dt)
            last_dt = dt

    def test_next_previous_refill(self):
        for appointment in Appointment.objects.all().order_by("timepoint"):
            subject_visit = SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=appointment.appt_datetime,
                reason=SCHEDULED,
            )
            if appointment.next:
                StudyMedication.objects.create(
                    subject_visit=subject_visit,
                    report_datetime=subject_visit.report_datetime,
                    refill_start_datetime=subject_visit.report_datetime,
                    refill_end_datetime=get_next_appointment(
                        appointment, include_interim=True
                    ).appt_datetime
                    - relativedelta(minutes=1),
                    dosage_guideline=self.dosage_guideline_100,
                    formulation=self.formulation,
                )
        obj0 = StudyMedication.objects.all().order_by("refill_start_datetime")[0]
        obj1 = StudyMedication.objects.all().order_by("refill_start_datetime")[1]
        self.assertEqual(obj0.next.id, obj1.id)
        self.assertEqual(obj0.id, obj1.previous.id)

    def test_insert_unscheduled_appt_refill(self):
        for appointment in Appointment.objects.all().order_by("timepoint"):
            subject_visit = SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=appointment.appt_datetime,
                reason=SCHEDULED,
            )
            if appointment.next:
                StudyMedication.objects.create(
                    subject_visit=subject_visit,
                    report_datetime=subject_visit.report_datetime,
                    refill_start_datetime=subject_visit.report_datetime,
                    refill_end_datetime=get_next_appointment(
                        appointment, include_interim=True
                    ).appt_datetime
                    - relativedelta(minutes=1),
                    dosage_guideline=self.dosage_guideline_100,
                    formulation=self.formulation,
                )

        for appointment in Appointment.objects.all():
            appointment.appt_status = INCOMPLETE_APPT
            appointment.save_base(update_fields=["appt_status"])

        prev_obj = None
        for obj in StudyMedication.objects.all().order_by("refill_start_datetime"):
            if not prev_obj:
                prev_obj = obj
                continue
        appointment_before = Appointment.objects.all().order_by("timepoint")[1]
        appointment_after = Appointment.objects.all().order_by("timepoint")[2]
        creator = UnscheduledAppointmentCreator(
            subject_identifier=appointment_before.subject_identifier,
            visit_schedule_name=appointment_before.visit_schedule_name,
            schedule_name=appointment_before.schedule_name,
            visit_code=appointment_before.visit_code,
            facility=appointment_before.facility,
            timepoint=appointment_before.timepoint,
        )
        subject_visit = SubjectVisit.objects.create(
            appointment=creator.appointment,
            report_datetime=creator.appointment.appt_datetime,
            reason=UNSCHEDULED,
        )
        study_medication = StudyMedication(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            refill_start_datetime=subject_visit.report_datetime,
            refill_end_datetime=appointment_after.appt_datetime,
            dosage_guideline=self.dosage_guideline_100,
            formulation=self.formulation,
        )
        study_medication.save()

        self.assertEqual(Appointment.objects.all().count(), 4)
        self.assertEqual(RxRefill.objects.all().count(), 3)

        prev_obj = None
        for obj in StudyMedication.objects.all().order_by("refill_start_datetime"):
            if not prev_obj:
                prev_obj = obj
                continue
            self.assertLess(prev_obj.refill_start_datetime, obj.refill_start_datetime)
            self.assertLess(prev_obj.refill_end_datetime, obj.refill_start_datetime)
            self.assertLess(prev_obj.refill_end_datetime, obj.refill_end_datetime)
            prev_obj = obj

        prev_obj = None
        for obj in RxRefill.objects.all().order_by("refill_start_datetime"):
            if not prev_obj:
                prev_obj = obj
                continue
            self.assertLess(prev_obj.refill_start_datetime, obj.refill_start_datetime)
            self.assertLess(prev_obj.refill_end_datetime, obj.refill_start_datetime)
            self.assertLess(prev_obj.refill_end_datetime, obj.refill_end_datetime)
            prev_obj = obj

    def test_for_all_appts(self):
        """Assert for all appointments.

        Captures exception at last appointment where "next" is none
        """
        for appointment in Appointment.objects.all().order_by("timepoint"):
            subject_visit = SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=appointment.appt_datetime,
                reason=SCHEDULED,
            )
            if not appointment.next:
                self.assertRaises(
                    NextStudyMedicationError,
                    StudyMedication.objects.create,
                    subject_visit=subject_visit,
                    report_datetime=subject_visit.report_datetime,
                    refill_start_datetime=subject_visit.report_datetime,
                    refill_end_datetime=None,
                    dosage_guideline=self.dosage_guideline_100,
                    formulation=self.formulation,
                )
            else:
                StudyMedication.objects.create(
                    subject_visit=subject_visit,
                    report_datetime=subject_visit.report_datetime,
                    refill_start_datetime=subject_visit.report_datetime,
                    refill_end_datetime=get_next_appointment(
                        subject_visit.appointment, include_interim=True
                    ).appt_datetime,
                    dosage_guideline=self.dosage_guideline_100,
                    formulation=self.formulation,
                )

    def test_study_medication_form_baseline(self):
        self.study_open_datetime = Protocol().study_open_datetime
        self.study_close_datetime = Protocol().study_close_datetime
        consent_object_factory(
            model="edc_pharmacy.subjectconsent",
            start=self.study_open_datetime,
            end=self.study_close_datetime,
        )

        appointment = Appointment.objects.all().order_by("timepoint")[0]
        next_appointment = get_next_appointment(appointment, include_interim=True)
        subject_visit = SubjectVisit.objects.create(
            appointment=appointment,
            report_datetime=appointment.appt_datetime,
            reason=SCHEDULED,
        )
        data = dict(
            subject_visit=subject_visit,
            refill=YES,
            report_datetime=subject_visit.report_datetime,
            refill_start_datetime=subject_visit.report_datetime,
            refill_end_datetime=next_appointment.appt_datetime,
            dosage_guideline=self.dosage_guideline_100,
            formulation=self.formulation,
            refill_to_next_visit=YES,
            roundup_divisible_by=32,
        )

        form = StudyMedicationForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)

    def test_inserts_refill(self):
        # 1000
        appointment = Appointment.objects.all().order_by("timepoint")[0]
        subject_visit = SubjectVisit.objects.create(
            appointment=appointment,
            report_datetime=appointment.appt_datetime,
            reason=SCHEDULED,
        )
        self.assertEqual(RxRefill.objects.all().count(), 0)
        StudyMedication.objects.create(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            refill_start_datetime=subject_visit.report_datetime,
            refill_end_datetime=get_next_appointment(
                appointment, include_interim=True
            ).appt_datetime,
            dosage_guideline=self.dosage_guideline_100,
            formulation=self.formulation,
        )
        self.assertEqual(RxRefill.objects.all().count(), 1)
        refills = RxRefill.objects.all().order_by("refill_start_datetime")
        self.assertEqual(refills[0].dosage_guideline, self.dosage_guideline_100)

        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()

        # 2000
        appointment = Appointment.objects.all().order_by("timepoint")[1]
        subject_visit = SubjectVisit.objects.create(
            appointment=appointment,
            report_datetime=appointment.appt_datetime,
            reason=SCHEDULED,
        )

        StudyMedication.objects.create(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            refill_start_datetime=subject_visit.report_datetime,
            refill_end_datetime=get_next_appointment(
                appointment, include_interim=True
            ).appt_datetime,
            dosage_guideline=self.dosage_guideline_200,
            formulation=self.formulation,
        )

        self.assertEqual(RxRefill.objects.all().count(), 2)

        refills = RxRefill.objects.all().order_by("refill_start_datetime")
        self.assertEqual(refills[0].dosage_guideline, self.dosage_guideline_100)
        self.assertEqual(refills[1].dosage_guideline, self.dosage_guideline_200)
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()

        # 3000
        appointment = Appointment.objects.all().order_by("timepoint")[2]
        subject_visit = SubjectVisit.objects.create(
            appointment=appointment,
            report_datetime=appointment.appt_datetime,
            reason=SCHEDULED,
        )

        opts = dict(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            refill_start_datetime=subject_visit.report_datetime,
            refill_end_datetime=None,
            dosage_guideline=self.dosage_guideline_200,
            formulation=self.formulation,
        )
        self.assertRaises(NextStudyMedicationError, StudyMedication.objects.create, **opts)
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()

        self.assertEqual(RxRefill.objects.all().count(), 2)
        refills = RxRefill.objects.all().order_by("refill_start_datetime")
        self.assertEqual(refills[0].dosage_guideline, self.dosage_guideline_100)
        self.assertEqual(refills[1].dosage_guideline, self.dosage_guideline_200)

        # insert unscheduled appt between 2000 and 3000

        appointment = Appointment.objects.all().order_by("timepoint")[1]
        creator = UnscheduledAppointmentCreator(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
            facility=appointment.facility,
            timepoint=appointment.timepoint,
        )

        subject_visit = SubjectVisit.objects.create(
            appointment=creator.appointment,
            report_datetime=creator.appointment.appt_datetime,
            reason=UNSCHEDULED,
        )

        StudyMedication.objects.create(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            refill_start_datetime=subject_visit.report_datetime,
            refill_end_datetime=get_next_appointment(
                creator.appointment, include_interim=True
            ).appt_datetime,
            dosage_guideline=self.dosage_guideline_100,
            formulation=self.formulation,
        )

        self.assertEqual(StudyMedication.objects.all().count(), 3)

        self.assertEqual(
            self.dosage_guideline_100,
            StudyMedication.objects.all()
            .order_by("refill_start_datetime")[0]
            .dosage_guideline,
        )
        self.assertEqual(
            self.dosage_guideline_200,
            StudyMedication.objects.all()
            .order_by("refill_start_datetime")[1]
            .dosage_guideline,
        )
        self.assertEqual(
            self.dosage_guideline_100,
            StudyMedication.objects.all()
            .order_by("refill_start_datetime")[2]
            .dosage_guideline,
        )



CANCELLED = "cancelled"
CONTROL_ARM = "control_arm"
DAYS = "days"
DISPENSED = "dispensed"
FILLED = "filled"
IN_PROGRESS_APPT = "in_progress"
PARTIAL = "partial"
PER_DAY = "day"
PER_HR = "hr"
PRESCRIPTION_ACTION = "prescription"
PRINT_SELECTED = "print_selected"
SINGLE = "single"
SOLUTION = "SOLUTION"



from typing import Tuple

from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse
from django_audit_fields.admin import audit_fieldset_tuple

from ..admin_site import edc_pharmacy_admin
from ..forms import DispensingHistoryForm
from ..models import DispensingHistory
from .model_admin_mixin import ModelAdminMixin


@admin.register(DispensingHistory, site=edc_pharmacy_admin)
class DispensingHistoryAdmin(ModelAdminMixin, admin.ModelAdmin):

    form = DispensingHistoryForm

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "rx_refill",
                    "dispensed",
                    "status",
                    "dispensed_datetime",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display: Tuple[str, ...] = (
        "subject_identifier",
        "refill",
        "description",
        "dispensed",
        "dispensed_date",
    )
    list_filter: Tuple[str, ...] = ("dispensed_datetime", "status")
    search_fields: Tuple[str, ...] = (
        "rx_refill__id",
        "rx_refill__rx__subject_identifier",
        "rx_refill__dosage_guideline__medication__name",
    )
    ordering: Tuple[str, ...] = ("dispensed_datetime",)

    @admin.display(description="Subject identifier")
    def subject_identifier(self, obj=None):
        return obj.rx_refill.rx.subject_identifier

    @admin.display(description="Refill")
    def refill(self, obj=None):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_rxrefill_changelist")
        url = f"{url}?q={obj.rx_refill.id}"
        context = dict(title="Back to RX refill", url=url, label="Refill")
        return render_to_string("dashboard_button.html", context=context)

    @admin.display(description="description")
    def description(self, obj=None):
        return obj.rx_refill


class DispensingHistoryInlineAdmin(admin.TabularInline):
    def has_add_permission(self, request, obj):
        return False

    form = DispensingHistoryForm
    model = DispensingHistory
    can_delete = False

    fields = ("dispensed", "status", "dispensed_datetime")
    ordering = ("dispensed_datetime",)
    readonly_fields = ("dispensed", "status", "dispensed_datetime")
    extra = 0



from edc_reference import site_reference_configs

site_reference_configs.register_from_visit_schedule(
    visit_models={"edc_appointment.appointment": "edc_pharmacy.subjectvisit"}
)

configs = {
    "edc_pharmacy.studymedication": ["report_datetime"],
}

for reference_name, fields in configs.items():
    site_reference_configs.add_fields_to_config(name=reference_name, fields=fields)



class ActivePrescriptionRefillOverlap(Exception):
    pass


class RefillError(Exception):
    pass


class ActiveRefillAlreadyExists(Exception):
    pass


class NextRefillError(Exception):
    pass


class NextStudyMedicationError(Exception):
    pass


class InsufficientQuantityError(Exception):
    pass


class PackagingSidMismatchError(Exception):
    pass


class PackagingSubjectIdentifierMismatchError(Exception):
    pass


class PackagingLotNumberMismatchError(Exception):
    pass


class PrescriptionAlreadyExists(Exception):
    pass


class PrescriptionError(Exception):
    pass


class PrescriptionExpired(Exception):
    pass


class RefillAlreadyExists(Exception):
    pass


class PrescriptionNotStarted(Exception):
    pass


class RefillCreatorError(Exception):
    pass


class RefillEndDatetimeError(Exception):
    pass


class StudyMedicationError(Exception):
    pass



from uuid import uuid4

from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from ..dispense import Dispensing
from .dispensing_history import DispensingHistory

# from .order import Order
# from .rx_refill import RxRefill
from .stock_create_labels import Labels, StockCreateLabels


@receiver(post_save, sender=DispensingHistory, dispatch_uid="dispensing_history_on_post_save")
def dispensing_history_on_post_save(sender, instance, raw, created, **kwargs):
    if not raw:
        dispensing = Dispensing(rx_refill=instance.rx_refill, dispensed=instance.dispensed)
        instance.rx_refill.remaining = dispensing.remaining
        instance.rx_refill.save(update_fields=["remaining"])


@receiver(
    post_delete,
    sender=DispensingHistory,
    dispatch_uid="dispensing_history_on_post_delete",
)
def dispensing_history_on_post_delete(sender, instance, using=None, **kwargs):
    dispensing = Dispensing(rx_refill=instance.rx_refill, dispensed=instance.dispensed)
    instance.rx_refill.remaining = dispensing.remaining
    instance.rx_refill.save(update_fields=["remaining"])


@receiver(
    post_save,
    dispatch_uid="create_or_update_refills_on_post_save",
)
def create_or_update_refills_on_post_save(
    sender, instance, raw, created, update_fields, **kwargs
):
    if not raw:
        try:
            instance.related_visit_model_attr()  # see edc-visit-tracking
        except AttributeError:
            pass
        else:
            try:
                instance.creates_refills_from_crf()
            except AttributeError as e:
                if "creates_refills_from_crf" not in str(e):
                    raise
                pass


# @receiver(
#     post_save,
#     sender=RxRefill,
#     dispatch_uid="create_or_update_refills_on_post_save",
# )
# def get_or_create_subject_orders_on_refill(
#     sender, instance, raw, created, update_fields, **kwargs
# ):
#     """Confirm or create an order for the items requested thru the refill"""
#     if not raw:
#         pass
#         instance.total
#         Order.objects.filter(product=product)
#         # check orders
#         # is the product requested on order
#         # is the product requested on order and delivered to
#         # check patient_stock


@receiver(
    post_save,
    sender=StockCreateLabels,
    dispatch_uid="create_stock_labels_on_post_save",
)
def create_stock_labels_on_post_save(sender, instance, raw, created, **kwargs):
    if not raw:
        qty_already_created = Labels.objects.filter(stock_create_labels=instance).count()
        for i in range(0, instance.qty - qty_already_created):
            Labels.objects.create(stock_create_labels=instance, stock_identifier=uuid4().hex)



from django.core.validators import MinValueValidator
from django.db import models
from django.db.models import PROTECT
from edc_model import models as edc_models
from edc_utils.round_up import round_half_away_from_zero

from .list_models import FrequencyUnits, Units
from .medication import Medication


class Manager(models.Manager):

    use_in_migrations = True

    def get_by_natural_key(self, medication_name, dose, dose_units, dose_per_kg):
        return self.get(medication_name, dose, dose_units, dose_per_kg)


class DosageGuideline(edc_models.BaseUuidModel):

    """Dosage guidelines."""

    medication = models.ForeignKey(Medication, on_delete=PROTECT, null=True, blank=False)

    dose = models.DecimalField(
        max_digits=8,
        decimal_places=2,
        null=True,
        blank=True,
        help_text="dose per 'frequency unit' if NOT considering subject's weight",
    )

    dose_per_kg = models.DecimalField(
        max_digits=8,
        decimal_places=2,
        null=True,
        blank=True,
        help_text="dose per 'frequency unit' if considering subject's weight",
    )

    dose_units = models.ForeignKey(Units, on_delete=PROTECT)

    frequency = models.DecimalField(
        verbose_name="Frequency",
        max_digits=6,
        decimal_places=2,
        validators=[MinValueValidator(1.0)],
        default=1,
        help_text="number of times per 'frequency unit'",
    )

    frequency_units = models.ForeignKey(
        FrequencyUnits,
        verbose_name="Frequency unit",
        on_delete=PROTECT,
    )

    objects = Manager()

    history = edc_models.HistoricalRecords()

    def __str__(self):
        return (
            f"{self.medication.name} {round_half_away_from_zero(self.dose or 0, 0)}"
            f"{self.dose_units} {round_half_away_from_zero((self.frequency or 0), 0)} "
            f"{self.get_frequency_units_display()}{' (per kg)' if self.dose_per_kg else ''}"
        )

    def natural_key(self):
        return (
            self.medication,
            self.dose,
            self.dose_units,
            self.dose_per_kg,
        )

    def get_dose_units_display(self):
        return self.dose_units.display_name

    def get_frequency_units_display(self):
        return self.frequency_units.display_name

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Dosage Guideline"
        verbose_name_plural = "Dosage Guidelines"
        unique_together = [
            ["medication", "dose", "dose_units", "dose_per_kg"],
        ]



from edc_constants.constants import NEW, NOT_APPLICABLE, OTHER

from .constants import CANCELLED, DISPENSED, FILLED, PARTIAL

PRESCRIPTION_STATUS = (
    (NEW, "New"),
    (PARTIAL, "Partially filled"),
    (FILLED, "Filled"),
    (CANCELLED, "Cancelled"),
)


DISPENSE_STATUS = ((DISPENSED, "Dispensed"), (CANCELLED, "Cancelled"))


FREQUENCY = (
    ("hr", "per hour"),
    ("day", "per day"),
    ("single", "single dose"),
    (OTHER, "Other ..."),
    (NOT_APPLICABLE, "Not applicable"),
)



from uuid import uuid4

from django.db import models
from django.db.models import PROTECT
from edc_model import models as edc_models
from edc_sites.models import SiteModelMixin

from .formulation import Formulation
from .list_models import Container
from .medication_lot import MedicationLot


class Manager(models.Manager):

    use_in_migrations = True


class Product(SiteModelMixin, edc_models.BaseUuidModel):

    product_identifier = models.CharField(max_length=36, default=uuid4, unique=True)

    name = models.CharField(max_length=250, unique=True, editable=False)

    container = models.ForeignKey(Container, on_delete=PROTECT)

    count_per_container = models.DecimalField(
        verbose_name="Items per container", max_digits=6, decimal_places=1
    )

    formulation = models.ForeignKey(Formulation, on_delete=PROTECT)

    medication_lot = models.ForeignKey(MedicationLot, on_delete=PROTECT)

    objects = Manager()

    history = edc_models.HistoricalRecords()

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        self.name = (
            f"{self.formulation.medication.display_name} "
            f"{self.formulation.strength}{self.formulation.units}. "
            f"LOT#: {self.medication_lot.lot_no}. {self.container} of "
            f"{self.count_per_container}"
        )
        super().save(*args, **kwargs)

    class Meta(SiteModelMixin.Meta, edc_models.BaseUuidModel.Meta):
        verbose_name = "Medication product"
        verbose_name_plural = "Medication products"



from django.db import models
from django.db.models.deletion import PROTECT
from edc_model import models as edc_models
from edc_utils import get_utcnow

from .rx_refill import RxRefill


class ReturnError(Exception):
    pass


class Manager(models.Manager):

    use_in_migrations = True

    def get_by_natural_key(self, prescription_item, return_datetime):
        return self.get(prescription_item, return_datetime)


class ReturnHistory(edc_models.BaseUuidModel):

    rx_refill = models.ForeignKey(RxRefill, on_delete=PROTECT)

    return_datetime = models.DateTimeField(default=get_utcnow)

    returned = models.DecimalField(max_digits=6, decimal_places=1)

    objects = Manager()

    history = edc_models.HistoricalRecords()

    def __str__(self):
        return f"{str(self.rx_refill)}"

    def natural_key(self):
        return (
            self.rx_refill,
            self.return_datetime,
        )

    # TODO: calculate to verify number of returns makes sense
    # def save(self, *args, **kwargs):
    #     if self.prescription_item.get_remaining(exclude_id=self.id) < self.returned:
    #         raise ReturnError("Attempt to return more than prescribed.")
    #     super().save(*args, **kwargs)

    @property
    def return_date(self):
        return self.return_datetime.date()

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Return history"
        verbose_name_plural = "Return history"
        unique_together = ["rx_refill", "return_datetime"]



# Generated by Django 3.2.13 on 2022-07-08 05:13

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0009_auto_20220708_0805"),
    ]

    operations = [
        migrations.RenameField(
            model_name="historicallabels",
            old_name="medication_stock_create_labels",
            new_name="stock_create_labels",
        ),
        migrations.RenameField(
            model_name="historicalrxrefill",
            old_name="medication_stock",
            new_name="stock",
        ),
        migrations.RenameField(
            model_name="historicalstock",
            old_name="medication_product",
            new_name="product",
        ),
        migrations.RenameField(
            model_name="historicalstockcreatelabels",
            old_name="medication_product",
            new_name="product",
        ),
        migrations.RenameField(
            model_name="historicalstockreceiving",
            old_name="medication_product",
            new_name="product",
        ),
        migrations.RenameField(
            model_name="labels",
            old_name="medication_stock_create_labels",
            new_name="stock_create_labels",
        ),
        migrations.RenameField(
            model_name="rxrefill",
            old_name="medication_stock",
            new_name="stock",
        ),
        migrations.RenameField(
            model_name="stock",
            old_name="medication_product",
            new_name="product",
        ),
        migrations.RenameField(
            model_name="stockcreatelabels",
            old_name="medication_product",
            new_name="product",
        ),
        migrations.RenameField(
            model_name="stockreceiving",
            old_name="medication_product",
            new_name="product",
        ),
    ]



from dateutil.relativedelta import relativedelta
from edc_consent.utils import get_consent_model_name
from edc_visit_schedule import Crf, FormsCollection, Schedule, Visit, VisitSchedule

app_label = "edc_pharmacy"

crfs = FormsCollection(
    Crf(show_order=1, model=f"{app_label}.studymedication", required=True),
)

visit0 = Visit(
    code="1000",
    title="Week 1",
    timepoint=0,
    rbase=relativedelta(days=0),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs,
    allow_unscheduled=True,
    facility_name="5-day-clinic",
)

visit1 = Visit(
    code="2000",
    title="Week 2",
    timepoint=1,
    rbase=relativedelta(days=7),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs,
    facility_name="5-day-clinic",
    allow_unscheduled=True,
)

visit2 = Visit(
    code="3000",
    title="Week 3",
    timepoint=2,
    rbase=relativedelta(days=14),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs,
    facility_name="5-day-clinic",
)

schedule = Schedule(
    name="schedule",
    onschedule_model="edc_pharmacy.onschedule",
    offschedule_model="edc_pharmacy.offschedule",
    consent_model=get_consent_model_name(),
    appointment_model="edc_appointment.appointment",
)

schedule.add_visit(visit0)
schedule.add_visit(visit1)
schedule.add_visit(visit2)

visit_schedule = VisitSchedule(
    name="visit_schedule",
    offstudy_model="edc_offstudy.subjectoffstudy",
    death_report_model="edc_adverse_event.deathreport",
)

visit_schedule.add_schedule(schedule)



from dateutil.relativedelta import relativedelta
from django.test import TestCase
from edc_constants.constants import FEMALE
from edc_registration.models import RegisteredSubject
from edc_utils import get_utcnow

from edc_pharmacy.exceptions import PrescriptionNotStarted
from edc_pharmacy.models import (
    DosageGuideline,
    Formulation,
    FormulationType,
    FrequencyUnits,
    Medication,
    Route,
    Rx,
    RxRefill,
    Units,
)
from edc_pharmacy.refill import (
    RefillCreator,
    activate_refill,
    deactivate_refill,
    get_active_refill,
)


class TestRefill(TestCase):
    def setUp(self):
        self.subject_identifier = "12345"

        report_datetime = get_utcnow() - relativedelta(days=60)
        RegisteredSubject.objects.create(
            subject_identifier="12345",
            dob=get_utcnow() - relativedelta(years=25),
            gender=FEMALE,
            registration_datetime=report_datetime,
            randomization_datetime=report_datetime,
        )

        self.medication = Medication.objects.create(
            name="Flucytosine",
            display_name="flucytosine",
        )

        # create prescription for this subject
        self.rx = Rx.objects.create(
            subject_identifier=self.subject_identifier,
            weight_in_kgs=40,
            report_datetime=report_datetime,
            rx_date=report_datetime.date(),
        )
        self.rx.medications.add(self.medication)

        self.formulation = Formulation.objects.create(
            medication=self.medication,
            strength=500,
            units=Units.objects.get(name="mg"),
            route=Route.objects.get(display_name="Oral"),
            formulation_type=FormulationType.objects.get(display_name__iexact="Tablet"),
        )

        self.dosage_guideline = DosageGuideline.objects.create(
            medication=self.medication,
            dose_per_kg=50,
            dose_units=Units.objects.get(name="mg"),
            frequency=1,
            frequency_units=FrequencyUnits.objects.get(name="day"),
        )

        self.dosage_guideline_no_weight = DosageGuideline.objects.create(
            medication=self.medication,
            dose=500,
            dose_units=Units.objects.get(name="mg"),
            frequency=2,
            frequency_units=FrequencyUnits.objects.get(name="day"),
        )

    def test_rx_refill_str(self):
        obj = RxRefill.objects.create(
            rx=self.rx,
            formulation=self.formulation,
            dosage_guideline=self.dosage_guideline,
            refill_start_datetime=get_utcnow(),
            refill_end_datetime=get_utcnow() + relativedelta(days=10),
            weight_in_kgs=65,
        )
        self.assertTrue(str(obj))

    def test_prescription_calculates_dose(self):
        """50mg /kg for 10 days = 65 * 50 * 10"""
        obj = RxRefill.objects.create(
            rx=self.rx,
            formulation=self.formulation,
            dosage_guideline=self.dosage_guideline,
            refill_start_datetime=get_utcnow(),
            refill_end_datetime=get_utcnow() + relativedelta(days=10),
            weight_in_kgs=65,
        )
        self.assertEqual(obj.dose, 6.5)
        self.assertEqual(obj.total, 65.0)  # 10 days
        self.assertEqual(obj.dosage_guideline.dose_units.name, "mg")
        self.assertEqual(obj.formulation.units.name, "mg")

    def test_prescription_total_to_dispense_or_order_without_roundup(self):
        obj = RxRefill.objects.create(
            rx=self.rx,
            formulation=self.formulation,
            dosage_guideline=self.dosage_guideline_no_weight,
            refill_start_datetime=get_utcnow(),
            refill_end_datetime=get_utcnow() + relativedelta(days=56),
        )
        self.assertEqual(obj.total, 112)

    def test_prescription_total_to_dispense_or_order_with_roundup(self):
        obj = RxRefill.objects.create(
            rx=self.rx,
            formulation=self.formulation,
            dosage_guideline=self.dosage_guideline_no_weight,
            refill_start_datetime=get_utcnow(),
            refill_end_datetime=get_utcnow() + relativedelta(days=56),
            roundup_divisible_by=32,
        )
        self.assertEqual(obj.total, 128)

    def test_prescription_total_to_dispense_or_order_weight_in_kgs(self):
        obj = RxRefill.objects.create(
            rx=self.rx,
            formulation=self.formulation,
            dosage_guideline=self.dosage_guideline,
            refill_start_datetime=get_utcnow(),
            refill_end_datetime=get_utcnow() + relativedelta(days=10),
            weight_in_kgs=65,
        )
        self.assertEqual(obj.dose, 6.5)  # 65kg * 50mg / 500mg = 6.5 pills
        self.assertEqual(obj.total, 65)  # x 10 days

    def test_prescription_total_to_dispense_or_order_weight_in_kgs_with_roundup(self):
        obj = RxRefill.objects.create(
            rx=self.rx,
            formulation=self.formulation,
            dosage_guideline=self.dosage_guideline,
            refill_start_datetime=get_utcnow(),
            refill_end_datetime=get_utcnow() + relativedelta(days=10),
            weight_in_kgs=65,
            roundup_divisible_by=12,
        )
        self.assertEqual(obj.dose, 6.5)  # 65kg * 50mg / 500mg = 6.5 pills
        self.assertEqual(obj.total, 72)

    def test_refill_gets_rx(self):
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            refill_start_datetime=get_utcnow(),
            refill_end_datetime=get_utcnow() + relativedelta(days=32),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            weight_in_kgs=65,
        )
        self.assertTrue(refill_creator.rx_refill.rx)

    def test_refill_raises_on_gets_rx(self):
        """Assert raises if refill date before Rx"""
        self.assertRaises(
            PrescriptionNotStarted,
            RefillCreator,
            subject_identifier=self.subject_identifier,
            refill_start_datetime=(get_utcnow() - relativedelta(years=1)),
            refill_end_datetime=get_utcnow() + relativedelta(days=32),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            weight_in_kgs=65,
        )

    def test_refill_create_and_no_active_refill(self):
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            refill_start_datetime=get_utcnow(),
            refill_end_datetime=get_utcnow() + relativedelta(days=32),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            make_active=False,
            weight_in_kgs=65,
        )
        self.assertIsNone(get_active_refill(refill_creator.rx_refill.rx))

    def test_refill_create_and_gets_active_refill(self):
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            refill_start_datetime=get_utcnow(),
            refill_end_datetime=get_utcnow() + relativedelta(days=32),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            weight_in_kgs=65,
        )
        self.assertEqual(
            get_active_refill(refill_creator.rx_refill.rx),
            refill_creator.rx_refill,
        )

    def test_refill_create_activates_by_default(self):
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            refill_start_datetime=get_utcnow(),
            refill_end_datetime=get_utcnow() + relativedelta(days=32),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            weight_in_kgs=65,
        )
        self.assertTrue(get_active_refill(refill_creator.rx_refill.rx).active)

    def test_refill_create_does_not_activate_if_false(self):
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            refill_start_datetime=get_utcnow(),
            refill_end_datetime=get_utcnow() + relativedelta(days=32),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            make_active=False,
            weight_in_kgs=65,
        )
        self.assertIsNone(get_active_refill(refill_creator.rx_refill.rx))

    def test_refill_create_duplicate_updates(self):
        # create a refill
        refill_start_datetime_one = get_utcnow() - relativedelta(days=60)
        RefillCreator(
            subject_identifier=self.subject_identifier,
            refill_start_datetime=refill_start_datetime_one,
            refill_end_datetime=refill_start_datetime_one + relativedelta(days=32),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            weight_in_kgs=65,
        )
        rx_refill_one = RxRefill.objects.get(
            rx__subject_identifier=self.subject_identifier,
            refill_start_datetime=refill_start_datetime_one,
        )
        self.assertEqual(rx_refill_one.number_of_days, 32)

        # create another refill but change number_of_days
        RefillCreator(
            subject_identifier=self.subject_identifier,
            refill_start_datetime=refill_start_datetime_one,
            refill_end_datetime=refill_start_datetime_one + relativedelta(days=31),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            weight_in_kgs=65,
        )
        rx_refill_two = RxRefill.objects.get(
            rx__subject_identifier=self.subject_identifier,
            refill_start_datetime=refill_start_datetime_one,
        )
        self.assertEqual(rx_refill_two.number_of_days, 31)
        self.assertEqual(rx_refill_two.id, rx_refill_one.id)

    def test_refill_create_finds_active(self):
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            refill_start_datetime=get_utcnow(),
            refill_end_datetime=get_utcnow() + relativedelta(days=32),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            make_active=True,
            weight_in_kgs=65,
        )
        self.assertIsNotNone(get_active_refill(refill_creator.rx_refill.rx))
        deactivate_refill(refill_creator.rx_refill)
        self.assertIsNone(get_active_refill(refill_creator.rx_refill.rx))

    def test_refill_create_activates_next(self):
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            refill_start_datetime=get_utcnow(),
            refill_end_datetime=get_utcnow() + relativedelta(days=32),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            weight_in_kgs=65,
        )
        self.assertIsNotNone(get_active_refill(refill_creator.rx_refill.rx))

        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            refill_start_datetime=get_utcnow() + relativedelta(months=1),
            refill_end_datetime=get_utcnow() + relativedelta(days=32),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            make_active=True,
            weight_in_kgs=65,
        )
        self.assertIsNotNone(get_active_refill(refill_creator.rx_refill.rx))
        self.assertEqual(
            get_active_refill(refill_creator.rx_refill.rx), refill_creator.rx_refill
        )

    def test_refill_create_refill_start_datetime(self):
        refill_start_datetime = get_utcnow()
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            refill_start_datetime=refill_start_datetime,
            refill_end_datetime=get_utcnow() + relativedelta(days=32),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            weight_in_kgs=65,
        )

        self.assertEqual(refill_creator.rx_refill.refill_start_datetime, refill_start_datetime)

    def test_refill_create_and_make_active(self):
        refill_start_datetime = get_utcnow()
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            refill_start_datetime=refill_start_datetime,
            refill_end_datetime=get_utcnow() + relativedelta(days=32),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            make_active=False,
            weight_in_kgs=65,
        )
        self.assertFalse(refill_creator.rx_refill.active)
        activate_refill(refill_creator.rx_refill)
        self.assertTrue(refill_creator.rx_refill.active)
        self.assertEqual(RxRefill.objects.all().count(), 1)

    def test_refill_count(self):
        refill_start_datetime = get_utcnow()
        RefillCreator(
            subject_identifier=self.subject_identifier,
            refill_start_datetime=refill_start_datetime,
            refill_end_datetime=get_utcnow() + relativedelta(days=11),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            make_active=True,
            weight_in_kgs=65,
            roundup_divisible_by=0,
        )
        rx_refill = RxRefill.objects.get(rx__subject_identifier=self.subject_identifier)
        self.assertEqual(rx_refill.number_of_days, 11)
        self.assertEqual(rx_refill.dose, 6.5)  # (65 * 50)/500
        self.assertEqual(rx_refill.total, 71.5)  # 6.5 * 11
        self.assertEqual(rx_refill.remaining, 71.5)

    def test_refill_on_change(self):
        refill_start_datetime = get_utcnow()
        RefillCreator(
            subject_identifier=self.subject_identifier,
            refill_start_datetime=refill_start_datetime,
            refill_end_datetime=get_utcnow() + relativedelta(days=11),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            make_active=True,
            weight_in_kgs=65,
            roundup_divisible_by=0,
        )
        rx_refill = RxRefill.objects.get(rx__subject_identifier=self.subject_identifier)
        self.assertEqual(rx_refill.number_of_days, 11)
        refill_start_datetime = get_utcnow() + relativedelta(days=4)
        RefillCreator(
            subject_identifier=self.subject_identifier,
            refill_start_datetime=refill_start_datetime,
            refill_end_datetime=get_utcnow() + relativedelta(days=11),
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            make_active=True,
            weight_in_kgs=65,
            roundup_divisible_by=0,
        )
        self.assertEqual(rx_refill.number_of_days, 11)



from .box import Box
from .container_type import ContainerType
from .items import ContainerModelMixin, GenericContainer, PillBottle, SubjectPillBottle
from .location import Location
from .room import Room
from .shelf import Shelf
from .unit_type import UnitType
from .utils import get_location, get_room, get_shelf, repackage, repackage_for_subject



from django.db import models
from edc_model import models as edc_models


class Manager(models.Manager):

    use_in_migrations = True

    def get_by_natural_key(self, name):
        return self.get(name)


class Medication(edc_models.BaseUuidModel):

    name = models.CharField(max_length=35, unique=True)

    display_name = models.CharField(max_length=50, unique=True)

    notes = models.TextField(max_length=250, null=True, blank=True)

    objects = Manager()

    history = edc_models.HistoricalRecords()

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        self.name = self.name.strip().lower().replace(" ", "_")
        super().save(*args, **kwargs)

    def natural_key(self):
        return self.name

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Medication"
        verbose_name_plural = "Medications"
        unique_together = ["name", "display_name"]



from __future__ import annotations

from typing import TYPE_CHECKING, Tuple

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from edc_appointment.utils import get_next_appointment
from edc_constants.constants import YES

from ..exceptions import NextStudyMedicationError, StudyMedicationError
from ..refill import create_refills_from_crf
from .previous_model_mixin import PreviousNextModelMixin
from .study_medication_refill_model_mixin import StudyMedicationRefillModelMixin

if TYPE_CHECKING:
    from ..models import RxRefill


class StudyMedicationCrfModelMixin(PreviousNextModelMixin, StudyMedicationRefillModelMixin):

    """Declare with field subject_visit using a CRF model mixin"""

    def save(self, *args, **kwargs):

        if not self.formulation:
            raise StudyMedicationError(
                f"Formulation cannot be None. Perhaps catch this in the form. See {self}."
            )
        if not self.dosage_guideline:
            raise StudyMedicationError(
                f"Dosage guideline cannot be None. Perhaps catch this in the form. See {self}."
            )

        if self.refill_to_next_visit == YES and not get_next_appointment(
            self.related_visit.appointment, include_interim=True
        ):
            raise NextStudyMedicationError(
                "Cannot refill to next appointment. This subject has no future appointments. "
                f"Perhaps catch this in the form. See {self.related_visit}."
            )

        if not self.refill_end_datetime:
            if self.refill_to_next_visit == YES:
                self.refill_end_datetime = get_next_appointment(
                    self.related_visit.appointment, include_interim=True
                ).appt_datetime
            else:
                self.refill_end_datetime = self.refill_start_datetime

        self.adjust_end_datetimes()

        self.number_of_days = (self.refill_end_datetime - self.refill_start_datetime).days

        if not self.rx:
            raise StudyMedicationError(
                f"Prescription not found. Perhaps catch this in the form. See {self}. "
            )

        super().save(*args, **kwargs)

    def creates_refills_from_crf(self) -> Tuple[RxRefill, RxRefill | None]:
        """Attribute called in signal"""
        return create_refills_from_crf(self, self.related_visit_model_attr())

    def get_subject_identifier(self):
        return self.related_visit.subject_identifier

    @property
    def rx(self):
        try:
            rx = django_apps.get_model("edc_pharmacy.rx").objects.get(
                registered_subject__subject_identifier=self.related_visit.subject_identifier,
                medications__in=[self.formulation.medication],
                rx_date__lte=self.refill_start_datetime.date(),
            )
        except ObjectDoesNotExist:
            return None
        else:
            if (
                rx.rx_expiration_date
                and rx.rx_expiration_date < self.refill_end_datetime.date()
            ):
                raise StudyMedicationError(
                    f"Prescription is expired. Got {rx}. Perhaps catch this in the form. "
                    f"See {self}."
                )
        return rx

    class Meta(StudyMedicationRefillModelMixin.Meta):
        abstract = True



from django import forms

from ..dispense import DispenseError, Dispensing
from ..models import DispensingHistory


class DispensingHistoryForm(forms.ModelForm):
    def clean(self):
        try:
            Dispensing(
                self.cleaned_data.get("rx_refill"),
                dispensed=self.cleaned_data.get("dispensed"),
                exclude_id=self.cleaned_data.get("id"),
            )
        except DispenseError as e:
            raise forms.ValidationError(e)

    class Meta:
        model = DispensingHistory
        fields = "__all__"


class DispensingHistoryReadonlyForm(forms.ModelForm):

    count = forms.DecimalField(
        label="Count", widget=forms.TextInput(attrs={"readonly": "readonly"})
    )

    status = forms.CharField(
        label="Status",
        widget=forms.TextInput(attrs={"readonly": "readonly"}),
    )

    dispensed_datetime = forms.DateTimeField(
        label="Dispensed Datetime",
        widget=forms.TextInput(attrs={"readonly": "readonly"}),
    )

    class Meta:
        model = DispensingHistory
        exclude = ["medication"]



from typing import Any

from django.core.exceptions import ObjectDoesNotExist

from ..utils import get_rxrefill_model_cls


def get_active_refill(rx: Any) -> Any:
    """Returns the 'active' Refill instance or None
    for this prescription.

    This does not return a model instance.
    """
    try:
        rx_refill = get_rxrefill_model_cls().objects.get(rx=rx, active=True)
    except ObjectDoesNotExist:
        return None
    return rx_refill



# Generated by Django 3.2.13 on 2022-09-13 18:39

import _socket
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django_audit_fields.models.audit_model_mixin
import django_crypto_fields.fields.encrypted_char_field
import django_revision.revision_field
import edc_pharmacy.models.order
import edc_utils.date
import simple_history.models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("edc_pharmacy", "0014_auto_20220907_0207"),
    ]

    operations = [
        migrations.CreateModel(
            name="Box",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "box_identifier",
                    models.CharField(default=uuid.uuid4, max_length=36, unique=True),
                ),
                ("box_datetime", models.DateTimeField(default=edc_utils.date.get_utcnow)),
                ("name", models.CharField(max_length=25, unique=True)),
                ("description", models.TextField(null=True)),
            ],
            options={
                "verbose_name": "Box",
                "verbose_name_plural": "Boxes",
                "ordering": ("-modified", "-created"),
                "get_latest_by": "modified",
                "abstract": False,
                "default_permissions": ("add", "change", "delete", "view", "export", "import"),
            },
        ),
        migrations.CreateModel(
            name="ContainerType",
            fields=[
                (
                    "name",
                    models.CharField(
                        db_index=True,
                        help_text="This is the stored value, required",
                        max_length=250,
                        unique=True,
                        verbose_name="Stored value",
                    ),
                ),
                (
                    "plural_name",
                    models.CharField(max_length=250, null=True, verbose_name="Plural name"),
                ),
                (
                    "display_name",
                    models.CharField(
                        db_index=True,
                        help_text="(suggest 40 characters max.)",
                        max_length=250,
                        unique=True,
                        verbose_name="Name",
                    ),
                ),
                (
                    "display_index",
                    models.IntegerField(
                        db_index=True,
                        default=0,
                        help_text="Index to control display order if not alphabetical, not required",
                        verbose_name="display index",
                    ),
                ),
                (
                    "field_name",
                    models.CharField(
                        blank=True,
                        editable=False,
                        help_text="Not required",
                        max_length=25,
                        null=True,
                    ),
                ),
                ("version", models.CharField(default="1.0", editable=False, max_length=35)),
                ("id", models.AutoField(primary_key=True, serialize=False)),
            ],
            options={
                "ordering": ["display_index", "display_name"],
                "abstract": False,
                "default_permissions": ("add", "change", "delete", "view", "export", "import"),
            },
        ),
        migrations.CreateModel(
            name="GenericContainer",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "container_identifier",
                    models.CharField(default=uuid.uuid4, max_length=36, unique=True),
                ),
                (
                    "container_datetime",
                    models.DateTimeField(default=edc_utils.date.get_utcnow),
                ),
                ("name", models.CharField(max_length=25, unique=True)),
                ("description", models.TextField(null=True)),
                ("contains_uniquely_identifiable_items", models.BooleanField(default=True)),
            ],
            options={
                "verbose_name": "Item",
                "verbose_name_plural": "Items",
                "ordering": ("-modified", "-created"),
                "get_latest_by": "modified",
                "abstract": False,
                "default_permissions": ("add", "change", "delete", "view", "export", "import"),
            },
        ),
        migrations.CreateModel(
            name="HistoricalOrder",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                ("subject_identifier", models.CharField(max_length=50)),
                ("order_datetime", models.DateTimeField(verbose_name="Order date/time")),
                (
                    "qty_ordered",
                    models.DecimalField(decimal_places=2, max_digits=10, null=True),
                ),
                (
                    "qty_supplied",
                    models.DecimalField(decimal_places=2, max_digits=10, null=True),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4, editable=False, primary_key=True, serialize=False
                    ),
                ),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical Medication order",
                "verbose_name_plural": "historical Medication orders",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="Location",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "location_identifier",
                    models.CharField(default=uuid.uuid4, max_length=36, unique=True),
                ),
                ("location_datetime", models.DateTimeField(default=edc_utils.date.get_utcnow)),
                ("name", models.CharField(max_length=25, unique=True)),
                ("description", models.TextField(null=True)),
            ],
            options={
                "verbose_name": "Location",
                "verbose_name_plural": "Locations",
                "ordering": ("-modified", "-created"),
                "get_latest_by": "modified",
                "abstract": False,
                "default_permissions": ("add", "change", "delete", "view", "export", "import"),
            },
        ),
        migrations.CreateModel(
            name="Order",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("subject_identifier", models.CharField(max_length=50)),
                ("order_datetime", models.DateTimeField(verbose_name="Order date/time")),
                (
                    "qty_ordered",
                    models.DecimalField(decimal_places=2, max_digits=10, null=True),
                ),
                (
                    "qty_supplied",
                    models.DecimalField(decimal_places=2, max_digits=10, null=True),
                ),
            ],
            options={
                "verbose_name": "Medication order",
                "verbose_name_plural": "Medication orders",
                "ordering": ("-modified", "-created"),
                "get_latest_by": "modified",
                "abstract": False,
                "default_permissions": ("add", "change", "delete", "view", "export", "import"),
            },
            managers=[
                ("objects", edc_pharmacy.models.order.Manager()),
            ],
        ),
        migrations.CreateModel(
            name="PillBottle",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "container_identifier",
                    models.CharField(default=uuid.uuid4, max_length=36, unique=True),
                ),
                (
                    "container_datetime",
                    models.DateTimeField(default=edc_utils.date.get_utcnow),
                ),
                ("name", models.CharField(max_length=25, unique=True)),
                ("description", models.TextField(null=True)),
                ("contains_uniquely_identifiable_items", models.BooleanField(default=True)),
                ("max_unit_qty", models.IntegerField(default=0)),
                ("unit_qty", models.IntegerField(default=0)),
                ("unit_qty_out", models.IntegerField(default=0)),
            ],
            options={
                "verbose_name": "Pill Bottle",
                "verbose_name_plural": "Pill Bottles",
                "ordering": ("-modified", "-created"),
                "get_latest_by": "modified",
                "abstract": False,
                "default_permissions": ("add", "change", "delete", "view", "export", "import"),
            },
        ),
        migrations.CreateModel(
            name="Room",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "room_identifier",
                    models.CharField(default=uuid.uuid4, max_length=36, unique=True),
                ),
                ("room_datetime", models.DateTimeField(default=edc_utils.date.get_utcnow)),
                ("name", models.CharField(max_length=25, unique=True)),
                ("description", models.TextField(null=True)),
            ],
            options={
                "verbose_name": "Room",
                "verbose_name_plural": "Rooms",
                "ordering": ("-modified", "-created"),
                "get_latest_by": "modified",
                "abstract": False,
                "default_permissions": ("add", "change", "delete", "view", "export", "import"),
            },
        ),
        migrations.CreateModel(
            name="Shelf",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "shelf_identifier",
                    models.CharField(default=uuid.uuid4, max_length=36, unique=True),
                ),
                ("shelf_datetime", models.DateTimeField(default=edc_utils.date.get_utcnow)),
                ("name", models.CharField(max_length=25, unique=True)),
                ("description", models.TextField(null=True)),
            ],
            options={
                "verbose_name": "Shelf",
                "verbose_name_plural": "Shelves",
                "ordering": ("-modified", "-created"),
                "get_latest_by": "modified",
                "abstract": False,
                "default_permissions": ("add", "change", "delete", "view", "export", "import"),
            },
        ),
        migrations.CreateModel(
            name="SubjectPillBottle",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "container_identifier",
                    models.CharField(default=uuid.uuid4, max_length=36, unique=True),
                ),
                (
                    "container_datetime",
                    models.DateTimeField(default=edc_utils.date.get_utcnow),
                ),
                ("name", models.CharField(max_length=25, unique=True)),
                ("description", models.TextField(null=True)),
                ("contains_uniquely_identifiable_items", models.BooleanField(default=True)),
                ("max_unit_qty", models.IntegerField(default=0)),
                ("unit_qty", models.IntegerField(default=0)),
                ("unit_qty_out", models.IntegerField(default=0)),
                ("rando_sid", models.CharField(max_length=25)),
                ("subject_identifier", models.CharField(max_length=50, null=True)),
            ],
            options={
                "verbose_name": "Subject Pill Bottle",
                "verbose_name_plural": "Subject Pill Bottles",
                "ordering": ("-modified", "-created"),
                "get_latest_by": "modified",
                "abstract": False,
                "default_permissions": ("add", "change", "delete", "view", "export", "import"),
            },
        ),
        migrations.CreateModel(
            name="UnitType",
            fields=[
                (
                    "name",
                    models.CharField(
                        db_index=True,
                        help_text="This is the stored value, required",
                        max_length=250,
                        unique=True,
                        verbose_name="Stored value",
                    ),
                ),
                (
                    "plural_name",
                    models.CharField(max_length=250, null=True, verbose_name="Plural name"),
                ),
                (
                    "display_name",
                    models.CharField(
                        db_index=True,
                        help_text="(suggest 40 characters max.)",
                        max_length=250,
                        unique=True,
                        verbose_name="Name",
                    ),
                ),
                (
                    "display_index",
                    models.IntegerField(
                        db_index=True,
                        default=0,
                        help_text="Index to control display order if not alphabetical, not required",
                        verbose_name="display index",
                    ),
                ),
                (
                    "field_name",
                    models.CharField(
                        blank=True,
                        editable=False,
                        help_text="Not required",
                        max_length=25,
                        null=True,
                    ),
                ),
                ("version", models.CharField(default="1.0", editable=False, max_length=35)),
                ("id", models.AutoField(primary_key=True, serialize=False)),
            ],
            options={
                "ordering": ["display_index", "display_name"],
                "abstract": False,
                "default_permissions": ("add", "change", "delete", "view", "export", "import"),
            },
        ),
        migrations.DeleteModel(
            name="MedicationOrder",
        ),
        migrations.RenameField(
            model_name="historicalproduct",
            old_name="lot_no",
            new_name="medication_lot",
        ),
        migrations.RenameField(
            model_name="product",
            old_name="lot_no",
            new_name="medication_lot",
        ),
        migrations.AddField(
            model_name="container",
            name="plural_name",
            field=models.CharField(max_length=250, null=True, verbose_name="Plural name"),
        ),
        migrations.AddField(
            model_name="formulationtype",
            name="plural_name",
            field=models.CharField(max_length=250, null=True, verbose_name="Plural name"),
        ),
        migrations.AddField(
            model_name="frequencyunits",
            name="plural_name",
            field=models.CharField(max_length=250, null=True, verbose_name="Plural name"),
        ),
        migrations.AddField(
            model_name="historicalmedicationlot",
            name="assignment",
            field=django_crypto_fields.fields.encrypted_char_field.EncryptedCharField(
                blank=True, help_text=" (Encryption: RSA local)", max_length=71, null=True
            ),
        ),
        migrations.AddField(
            model_name="medicationlot",
            name="assignment",
            field=django_crypto_fields.fields.encrypted_char_field.EncryptedCharField(
                blank=True, help_text=" (Encryption: RSA local)", max_length=71, null=True
            ),
        ),
        migrations.AddField(
            model_name="route",
            name="plural_name",
            field=models.CharField(max_length=250, null=True, verbose_name="Plural name"),
        ),
        migrations.AddField(
            model_name="units",
            name="plural_name",
            field=models.CharField(max_length=250, null=True, verbose_name="Plural name"),
        ),
        migrations.AddIndex(
            model_name="unittype",
            index=models.Index(
                fields=["id", "display_name", "display_index"],
                name="edc_pharmac_id_8faa96_idx",
            ),
        ),
        migrations.AddField(
            model_name="subjectpillbottle",
            name="box",
            field=models.ForeignKey(
                null=True, on_delete=django.db.models.deletion.PROTECT, to="edc_pharmacy.box"
            ),
        ),
        migrations.AddField(
            model_name="subjectpillbottle",
            name="container_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="edc_pharmacy.containertype"
            ),
        ),
        migrations.AddField(
            model_name="subjectpillbottle",
            name="formulation",
            field=models.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_pharmacy.formulation",
            ),
        ),
        migrations.AddField(
            model_name="subjectpillbottle",
            name="medication_lot",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="edc_pharmacy.medicationlot"
            ),
        ),
        migrations.AddField(
            model_name="subjectpillbottle",
            name="source_container",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_pharmacy.pillbottle",
            ),
        ),
        migrations.AddField(
            model_name="shelf",
            name="room",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="edc_pharmacy.room"
            ),
        ),
        migrations.AddField(
            model_name="room",
            name="location",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="edc_pharmacy.location"
            ),
        ),
        migrations.AddField(
            model_name="pillbottle",
            name="box",
            field=models.ForeignKey(
                null=True, on_delete=django.db.models.deletion.PROTECT, to="edc_pharmacy.box"
            ),
        ),
        migrations.AddField(
            model_name="pillbottle",
            name="container_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="edc_pharmacy.containertype"
            ),
        ),
        migrations.AddField(
            model_name="pillbottle",
            name="formulation",
            field=models.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_pharmacy.formulation",
            ),
        ),
        migrations.AddField(
            model_name="pillbottle",
            name="medication_lot",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="edc_pharmacy.medicationlot"
            ),
        ),
        migrations.AddField(
            model_name="pillbottle",
            name="source_container",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_pharmacy.pillbottle",
            ),
        ),
        migrations.AddField(
            model_name="order",
            name="product",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="edc_pharmacy.product"
            ),
        ),
        migrations.AddField(
            model_name="historicalorder",
            name="history_user",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="historicalorder",
            name="product",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_pharmacy.product",
            ),
        ),
        migrations.AddField(
            model_name="genericcontainer",
            name="box",
            field=models.ForeignKey(
                null=True, on_delete=django.db.models.deletion.PROTECT, to="edc_pharmacy.box"
            ),
        ),
        migrations.AddField(
            model_name="genericcontainer",
            name="container_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="edc_pharmacy.containertype"
            ),
        ),
        migrations.AddIndex(
            model_name="containertype",
            index=models.Index(
                fields=["id", "display_name", "display_index"],
                name="edc_pharmac_id_554046_idx",
            ),
        ),
        migrations.AddField(
            model_name="box",
            name="shelf",
            field=models.ForeignKey(
                null=True, on_delete=django.db.models.deletion.PROTECT, to="edc_pharmacy.shelf"
            ),
        ),
    ]



from django.db import models
from django.db.models import PROTECT
from edc_model import models as edc_models
from edc_utils.round_up import round_half_away_from_zero

from .list_models import FormulationType, Route, Units
from .medication import Medication


class Manager(models.Manager):

    use_in_migrations = True

    def get_by_natural_key(self, name, strength, units, formulation_type):
        return self.get(name, strength, units, formulation_type)


class Formulation(edc_models.BaseUuidModel):

    medication = models.ForeignKey(Medication, on_delete=PROTECT, null=True, blank=False)

    strength = models.DecimalField(max_digits=6, decimal_places=1)

    units = models.ForeignKey(Units, on_delete=PROTECT)

    formulation_type = models.ForeignKey(FormulationType, on_delete=PROTECT)

    route = models.ForeignKey(Route, on_delete=PROTECT)

    notes = models.TextField(max_length=250, null=True, blank=True)

    objects = Manager()

    history = edc_models.HistoricalRecords()

    def __str__(self):
        return self.description.title()

    def natural_key(self):
        return (
            self.medication,
            self.strength,
            self.units,
            self.formulation_type,
        )

    @property
    def description(self):
        return (
            f"{self.medication} {round_half_away_from_zero(self.strength, 0)}"
            f"{self.get_units_display()} "
            f"{self.get_formulation_type_display()} "
            f"{self.get_route_display()}"
        )

    def get_formulation_type_display(self):
        return self.formulation_type.display_name

    def get_units_display(self):
        return self.units.display_name

    def get_route_display(self):
        return self.route.display_name

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Formulation"
        verbose_name_plural = "Formulations"
        unique_together = ["medication", "strength", "units", "formulation_type"]



from .dispensing_history import DispensingHistory
from .dosage_guideline import DosageGuideline
from .formulation import Formulation
from .list_models import Container, FormulationType, FrequencyUnits, Route, Units
from .medication import Medication
from .medication_lot import MedicationLot
from .order import Order
from .product import Product
from .proxy_models import VisitSchedule
from .return_history import ReturnError, ReturnHistory
from .rx import Rx
from .rx_refill import RxRefill
from .signals import (
    create_or_update_refills_on_post_save,
    dispensing_history_on_post_save,
)
from .stock import Stock
from .stock_create_labels import Labels, StockCreateLabels
from .stock_receiving import StockReceiving
from .storage import (
    Box,
    ContainerModelMixin,
    ContainerType,
    GenericContainer,
    Location,
    PillBottle,
    Room,
    Shelf,
    SubjectPillBottle,
    UnitType,
    get_location,
    get_room,
    get_shelf,
    repackage,
    repackage_for_subject,
)
from .subject import Subject



from uuid import uuid4

from django.db import models
from django.db.models import PROTECT
from edc_model import models as edc_models

from .product import Product


class Manager(models.Manager):

    use_in_migrations = True


class StockCreateLabels(edc_models.BaseUuidModel):

    product = models.ForeignKey(Product, on_delete=PROTECT)

    qty = models.IntegerField(verbose_name="Number of labels to print")

    printed = models.BooleanField(default=False)

    printed_datetime = models.DateTimeField(null=True, blank=True)

    objects = Manager()

    history = edc_models.HistoricalRecords()

    def __str__(self):
        return f"{self.product}: {self.qty} "

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Medication stock: Create labels"
        verbose_name_plural = "Medication stock: Create labels"


class Labels(edc_models.BaseUuidModel):

    stock_create_labels = models.ForeignKey(StockCreateLabels, on_delete=PROTECT)

    stock_identifier = models.CharField(max_length=36, default=uuid4, unique=True)

    printed = models.BooleanField(default=False)

    printed_datetime = models.DateTimeField(null=True, blank=True)

    in_stock = models.BooleanField(default=False)

    in_stock_datetime = models.DateTimeField(null=True, blank=True)

    objects = Manager()

    history = edc_models.HistoricalRecords()

    def __str__(self):
        return f"{self.stock_labels}: {self.stock_identifier} "

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Label"
        verbose_name_plural = "Labels"



from django.db import models

from .pill_bottle import PillBottle
from .pill_bottle_model_mixin import PillBottleModelMixin


class SubjectPillBottle(PillBottleModelMixin):

    rando_sid = models.CharField(max_length=25)

    subject_identifier = models.CharField(max_length=50, null=True)

    source_container = models.ForeignKey(
        PillBottle, on_delete=models.PROTECT, blank=True, null=True
    )

    class Meta(PillBottleModelMixin.Meta):
        verbose_name = "Subject Pill Bottle"
        verbose_name_plural = "Subject Pill Bottles"



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from ..admin_site import edc_pharmacy_admin
from ..forms import LabelsForm
from ..models import Labels
from .actions import print_stock_labels
from .model_admin_mixin import ModelAdminMixin


@admin.register(Labels, site=edc_pharmacy_admin)
class LabelsAdmin(ModelAdminMixin, admin.ModelAdmin):

    show_object_tools = True

    form = LabelsForm

    actions = [print_stock_labels]

    fieldsets = (
        (
            None,
            {
                "fields": (
                    [
                        "stock_create_labels",
                        "stock_identifier",
                        "printed",
                        "printed_datetime",
                        "in_stock",
                        "in_stock_datetime",
                    ]
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "stock_create_labels",
        "stock_identifier",
        "printed",
        "printed_datetime",
        "in_stock",
        "in_stock_datetime",
        "created",
        "modified",
    )
    list_filter = (
        "stock_create_labels",
        "printed",
        "printed_datetime",
        "created",
        "modified",
    )
    search_fields = (
        "stock_create_labels__product__product_identifier",
        "stock_create_labels__product__lot_no",
    )
    ordering = ("printed_datetime",)
    readonly_fields = (
        "stock_create_labels",
        "stock_identifier",
        "printed",
        "printed_datetime",
        "in_stock",
        "in_stock_datetime",
    )



# Generated by Django 3.2.13 on 2022-07-04 15:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0006_historicalrxrefill_roundup_divisible_by_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="historicaldispensinghistory",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Dispensing history",
                "verbose_name_plural": "historical Dispensing history",
            },
        ),
        migrations.AlterModelOptions(
            name="historicaldosageguideline",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Dosage Guideline",
                "verbose_name_plural": "historical Dosage Guidelines",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalformulation",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Formulation",
                "verbose_name_plural": "historical Formulations",
            },
        ),
        migrations.AlterModelOptions(
            name="historicallabels",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Label",
                "verbose_name_plural": "historical Labels",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalmedication",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Medication",
                "verbose_name_plural": "historical Medications",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalmedicationlot",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Medication lot",
                "verbose_name_plural": "historical Medication lots",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalmedicationproduct",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Medication product",
                "verbose_name_plural": "historical Medication products",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalmedicationstock",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Medication stock",
                "verbose_name_plural": "historical Medication stock",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalmedicationstockcreatelabels",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Medication stock: Create labels",
                "verbose_name_plural": "historical Medication stock: Create labels",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalmedicationstockreceiving",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Medication stock: Receiving",
                "verbose_name_plural": "historical Medication stock: Receiving",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalreturnhistory",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Return history",
                "verbose_name_plural": "historical Return history",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalrx",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Prescription",
                "verbose_name_plural": "historical Prescriptions",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalrxrefill",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical RX refill",
                "verbose_name_plural": "historical RX refills",
            },
        ),
        migrations.AlterField(
            model_name="historicaldispensinghistory",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicaldosageguideline",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalformulation",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicallabels",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalmedication",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalmedicationlot",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalmedicationproduct",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalmedicationstock",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalmedicationstockcreatelabels",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalmedicationstockreceiving",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalreturnhistory",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalrx",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalrxrefill",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
    ]



from typing import Tuple

from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from ..admin_site import edc_pharmacy_admin
from ..forms import MedicationLotForm
from ..models import MedicationLot
from .model_admin_mixin import ModelAdminMixin


@admin.register(MedicationLot, site=edc_pharmacy_admin)
class MedicationLotAdmin(ModelAdminMixin, admin.ModelAdmin):

    show_object_tools = True

    form = MedicationLotForm

    fieldsets = (
        (
            None,
            {
                "fields": [
                    "lot_no",
                    "expiration_date",
                    "formulation",
                ]
            },
        ),
        audit_fieldset_tuple,
    )

    list_filter: Tuple[str, ...] = (
        "lot_no",
        "expiration_date",
        "formulation",
        "created",
        "modified",
    )

    list_display: Tuple[str, ...] = (
        "lot_no",
        "expiration_date",
        "formulation",
        "created",
        "modified",
    )

    search_fields: Tuple[str, ...] = ("lot_no",)

    ordering: Tuple[str, ...] = ("-expiration_date",)
